/*
    ext2.c -- generic ext2 stuff
    Copyright (C) 1998,99 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static const char _ext2_c[] = "$Id: ext2.c,v 1.46 1999/07/23 12:22:26 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "ext2.h"

#ifndef BLKGETSIZE
#define BLKGETSIZE _IO(0x12,96)
#endif

/* ext2 stuff ****************************************************************/

unsigned char _bitmap[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

void ext2_copy_block(struct ext2_fs *fs, blk_t from, blk_t to)
{
	unsigned char buf[fs->blocksize];

	ext2_bcache_flush(fs, from);
	ext2_bcache_flush(fs, to);

	ext2_read_blocks(fs, buf, from, 1);
	ext2_write_blocks(fs, buf, to, 1);
}

int ext2_get_block_state(struct ext2_fs *fs, blk_t block)
{
	struct ext2_buffer_head *bh;
	int group;
	int offset;
	int state;

	block -= fs->sb.s_first_data_block;
	group = block / fs->sb.s_blocks_per_group;
	offset = block % fs->sb.s_blocks_per_group;

	bh = ext2_bread(fs, fs->gd[group].bg_block_bitmap);
	state = bh->data[offset>>3] & _bitmap[offset&7];
	ext2_brelse(bh, 0);

	return state;
}

void ext2_read_blocks(struct ext2_fs *fs, void *ptr, blk_t block, blk_t numblocks)
{
	loff_t offset;

	offset = ((loff_t)block) << fs->logsize;
	if (fs->readoff != offset)
		llseek(fs->fdread, offset, SEEK_SET);

	read(fs->fdread, ptr, numblocks << fs->logsize);
	fs->readoff = ((loff_t)block+numblocks) << fs->logsize;
}

void ext2_set_block_state(struct ext2_fs *fs, blk_t block, int state, int updatemetadata)
{
	struct ext2_buffer_head *bh;
	int                      group;
	int                      offset;

	block -= fs->sb.s_first_data_block;
	group = block/fs->sb.s_blocks_per_group;
	offset = block%fs->sb.s_blocks_per_group;

	bh = ext2_bread(fs, fs->gd[group].bg_block_bitmap);
	bh->dirty = 1;
	if (state)
		bh->data[offset>>3] |= _bitmap[offset&7];
	else
		bh->data[offset>>3] &= ~_bitmap[offset&7];
	ext2_brelse(bh, 0);

	if (updatemetadata)
	{
		int diff;

		diff = state ? -1 : 1;

		fs->gd[group].bg_free_blocks_count += diff;
		fs->sb.s_free_blocks_count += diff;
		fs->metadirty = 1;
	}
}

void ext2_write_blocks(struct ext2_fs *fs, void *ptr, blk_t block, blk_t numblocks)
{
	loff_t offset;

	offset = ((loff_t)block) << fs->logsize;
	if (fs->writeoff != offset)
		llseek(fs->fdwrite, offset, SEEK_SET);

	write(fs->fdwrite, ptr, numblocks << fs->logsize);
	fs->writeoff = ((loff_t)block+numblocks) << fs->logsize;
}

void ext2_zero_block(struct ext2_fs *fs, blk_t block)
{
	struct ext2_buffer_head *bh;

	bh = ext2_bcreate(fs, block);
	bh->dirty = 1;
	ext2_brelse(bh, 1);
}





















off_t ext2_get_inode_offset(struct ext2_fs *fs, ino_t inode, blk_t *block)
{
	int group;
	int offset;

	inode--;

	group = inode / fs->sb.s_inodes_per_group;
	offset = (inode % fs->sb.s_inodes_per_group) * sizeof(struct ext2_inode);

	*block = fs->gd[group].bg_inode_table + (offset >> fs->logsize);

	return offset & (fs->blocksize - 1);
}

int ext2_get_inode_state(struct ext2_fs *fs, ino_t inode)
{
	struct ext2_buffer_head *bh;
	int                      group;
	int                      offset;
	int                      ret;

	inode--;
	group = inode / fs->sb.s_inodes_per_group;
	offset = inode % fs->sb.s_inodes_per_group;

	bh = ext2_bread(fs, fs->gd[group].bg_inode_bitmap);
	ret = bh->data[offset>>3] & _bitmap[offset&7];
	ext2_brelse(bh, 0);

	return ret;
}

void ext2_read_inode(struct ext2_fs *fs, ino_t inode, struct ext2_inode *data)
{
	struct ext2_buffer_head *bh;
	blk_t			 blk;
	off_t			 off;

	off = ext2_get_inode_offset(fs, inode, &blk);

	bh = ext2_bread(fs, blk);
	memcpy(data, bh->data + off, sizeof(struct ext2_inode));
	ext2_brelse(bh, 0);
}

void ext2_set_inode_state(struct ext2_fs *fs, ino_t inode, int state, int updatemetadata)
{
	struct ext2_buffer_head *bh;
	int                      group;
	int                      offset;

	inode--;
	group = inode / fs->sb.s_inodes_per_group;
	offset = inode % fs->sb.s_inodes_per_group;

	bh = ext2_bread(fs, fs->gd[group].bg_inode_bitmap);
	bh->dirty = 1;
	if (state)
		bh->data[offset>>3] |= _bitmap[offset&7];
	else
		bh->data[offset>>3] &= ~_bitmap[offset&7];
	ext2_brelse(bh, 0);

	if (updatemetadata)
	{
		int diff;

		diff = state ? -1 : 1;

		fs->gd[group].bg_free_inodes_count += diff;
		fs->sb.s_free_inodes_count += diff;
		fs->metadirty = 1;
	}
}

void ext2_write_inode(struct ext2_fs *fs, ino_t inode, const struct ext2_inode *data)
{
	struct ext2_buffer_head *bh;
	blk_t			 blk;
	off_t			 off;

	off = ext2_get_inode_offset(fs, inode, &blk);

	bh = ext2_bread(fs, blk);
	bh->dirty = 1;
	memcpy(bh->data + off, data, sizeof(struct ext2_inode));
	ext2_brelse(bh, 0);
}

void ext2_zero_inode(struct ext2_fs *fs, ino_t inode)
{
	struct ext2_inode buf;

	memset(&buf, 0, sizeof(struct ext2_inode));
	ext2_write_inode(fs, inode, &buf);
}





/* formula grabbed from linux ext2 kernel source */
static int is_root(int x, int y)
{
	if (!x)
		return 1;

	while (1)
	{
		if (x == 1)
			return 1;

		if (x % y)
			return 0;

		x /= y;
	}
}

int ext2_is_group_sparse(struct ext2_fs *fs, int group)
{
	if (!fs->sparse)
		return 1;

	if (is_root(group, 3) || is_root(group, 5) || is_root(group, 7))
		return 1;

	return 0;
}









void ext2_close(struct ext2_fs *fs)
{
	ext2_journal_deinit(fs);

	if (fs->metadirty)
		fs->sb.s_r_blocks_count = (fs->r_frac * (loff_t)fs->sb.s_blocks_count) / 100;

	ext2_commit_metadata(fs, 1);
	ext2_sync(fs);

	free(fs->relocator_pool);

	ext2_bcache_deinit(fs);

	close(fs->fdwrite);
	close(fs->fdread);

	free(fs->gd);
	free(fs);
}

void ext2_commit_metadata(struct ext2_fs *fs, int all_copies)
{
	int	      i;
	int	      num;
	unsigned char sb[fs->blocksize];
	int	      sboffset;

	if (!fs->metadirty || (!all_copies && fs->metadirty == 2))
		return;

	memset(sb, 0, fs->blocksize);

	sboffset = fs->sb.s_first_data_block;
	if (sboffset)
		memcpy(sb, &fs->sb, 1024);
	else
		memcpy(sb+1024, &fs->sb, 1024);

	num = 1;
	if (all_copies)
		num = fs->numgroups;

	i = 1;
	if (fs->metadirty == 1)
		i = 0;

	for (;i<num;i++)
	{
		struct ext2_buffer_head *bh;
		int                      sb_block;
		int                      j;

		if (!ext2_is_group_sparse(fs, i))
			continue;

		sb_block = sboffset + i * fs->sb.s_blocks_per_group;

		bh = ext2_bcreate(fs, sb_block);
		memcpy(bh->data, sb, fs->blocksize);
		ext2_brelse(bh, 1);

		for (j=0;j<fs->gdblocks;j++)
		{
			bh = ext2_bcreate(fs, sb_block + j + 1);
			memcpy(bh->data, ((unsigned char *)fs->gd) + (j << fs->logsize), fs->blocksize);
			ext2_brelse(bh, 1);
		}
	}

	if (all_copies)
		fs->metadirty = 0;
	else
		fs->metadirty = 2;
}

void ext2_sync(struct ext2_fs *fs)
{
	ext2_commit_metadata(fs, 0);
	ext2_bcache_sync(fs);
	fdatasync(fs->fdwrite);
}

struct ext2_fs *ext2_open(const char *name)
{
	struct ext2_fs *fs;
	struct stat st;

	if ((fs = (struct ext2_fs *)malloc(sizeof(struct ext2_fs))) == NULL)
	{
		fprintf(stderr, "ext2_open: MAerror!\n");
		goto err;
	}

	if ((fs->gd = (struct ext2_group_desc *)malloc(ext2_max_groups * sizeof(struct ext2_group_desc))) == NULL)
	{
		fprintf(stderr, "ext2_open: MAerror!\n");
		goto err1;
	}

	if ((fs->relocator_pool = (unsigned char *)malloc(ext2_relocator_pool_size << 10)) == NULL)
	{
		fprintf(stderr, "ext2_open: MAerror!\n");
		goto err2;
	}
	fs->relocator_pool_end = fs->relocator_pool + (ext2_relocator_pool_size << 10);

	if ((fs->fdread = open(name, O_RDONLY)) < 0)
	{
		perror("open");
		goto err3;
	}

	if ((fs->fdwrite = open(name, O_WRONLY)) < 0)
	{
		perror("dup");
		goto err4;
	}

	if (fstat(fs->fdread, &st) < 0)
	{
		perror("stat");
		goto err5;
	}

	fs->readoff = 0;
	fs->writeoff = 0;
	fs->logsize = 10;
	ext2_read_blocks(fs, &fs->sb, 1, 1);

	if (fs->sb.s_magic != EXT2_SUPER_MAGIC)
	{
		fprintf(stderr, "ext2_open: invalid superblock\n");
		goto err5;
	}

	if (fs->sb.s_state & EXT2_ERROR_FS)
	{
		fprintf(stderr, "ext2_open: filesystem has errors! e2fsck first!\n");
		goto err5;
	}

	if (!(fs->sb.s_state & EXT2_VALID_FS))
	{
		fprintf(stderr, "ext2_open: filesystem was not cleanly unmounted! e2fsck first!\n");
		goto err5;
	}

	if (fs->sb.s_feature_compat ||
	    fs->sb.s_feature_incompat ||
	    (fs->sb.s_feature_ro_compat & ~EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
	{
		fprintf(stderr, "ext2_open: fs has incompatible feature enabled\n");
		goto err5;
	}

	/* This is needed for the buffer cache initialization (enough
	   buffer heads must be allocated for the whole buffer cache
	   pool, and the number of buffer heads depends upon the size
	   of the blocks we divide the buffer cache pool in) */
	fs->logsize = fs->sb.s_log_block_size + 10;

	if (!ext2_bcache_init(fs))
	{
		fprintf(stderr, "ext2_open: MAerror!\n");
		goto err5;
	}

	/* From this point on all is well. No more memory allocations
	   are made */

	fs->blocksize = 1 << fs->logsize;

	fs->numgroups = howmany(fs->sb.s_blocks_count-fs->sb.s_first_data_block, fs->sb.s_blocks_per_group);
	fs->gdblocks = howmany(fs->numgroups * sizeof(struct ext2_group_desc), fs->blocksize);
	fs->inodeblocks = howmany(fs->sb.s_inodes_per_group * sizeof(struct ext2_inode), fs->blocksize);
	fs->r_frac = howmany(100 * (u_int64_t)fs->sb.s_r_blocks_count, fs->sb.s_blocks_count);
	fs->adminblocks = 3 + fs->gdblocks + fs->inodeblocks;

	fs->sparse = 0;
	if (fs->sb.s_feature_ro_compat & EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)
		fs->sparse = 1;

	ext2_read_blocks(fs, fs->gd, fs->sb.s_first_data_block + 1, fs->gdblocks);

	if (S_ISBLK(st.st_mode))      /* Thanks to Rolf Jakob */
	{
		if (ioctl(fs->fdread, BLKGETSIZE, &fs->containersize) < 0)
		{
			perror("ioctl");
			goto err5;
		}

		fs->containersize >>= fs->logsize - 9;
	}
	else
		fs->containersize = st.st_blocks >> (fs->logsize - 9);

	ext2_journal_init(fs);

	return fs;

 err5:
	close(fs->fdwrite);
 err4:
	close(fs->fdread);
 err3:
	free(fs->relocator_pool);
 err2:
	free(fs->gd);
 err1:
	free(fs);
 err:
	return NULL;
}
